#include "ARP2GTSP.h"
#include "Heap.h"

static Arc *CreateArc(int To, int Cost, int Required, Arc * Next);
static void Dijkstra(Vertex * Source);

static Vertex *V;
static int VSize;

/* The main function for ARP2GTSP, which transforms an ARP instance
 * into an GTSP instance.
 * Usage: ARP2GTSP ARP-class instance-name
 * The result of the transformation is placed in GTSP/instance-name.gtsp
 */

int main(int argc, char *argv[])
{
    int RequiredVertices = 0;
    int RequiredEdges = 0;
    int NotRequiredEdges = 0;
    int RequiredArcs = 0;
    int Edges = 0;
    int WSize, OldWSize;
    int Clusters;
    int i, j, x, y;
    Vertex *v, *v1, *v2;
    Wertex *W, *w;
    Arc *a;
    int **C;
    char *Class, Line[256];
    FILE *in, *out;

    Class = argv[1];
    sprintf(Line, "%s/%s", Class, argv[2] + (Class[0] == 'G' ? 4 : 0));

    /* Open instance file */
    if (!(in = fopen(Line, "r"))) {
        printf("No such file: %s\n", Line);
        exit(1);
    }
    printf("%s: \t", Line);
    fflush(stdout);

    /* Read instance file */
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "VERTICES", strlen("VERTICES")))
            break;
    sscanf(Line, "VERTICES :%d\n", &VSize);
    printf("|V| = %d, ", VSize);
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "ARISTAS_REQ", strlen("ARISTAS_REQ")))
            break;
    sscanf(Line, "ARISTAS_REQ :%d\n", &RequiredEdges);
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "ARISTAS_NOREQ", strlen("ARISTAS_NOREQ")))
            break;
    sscanf(Line, "ARISTAS_NOREQ :%d\n", &NotRequiredEdges);
    while (fgets(Line, 256, in))
        if (!strncmp(Line, "LISTA_A", strlen("LISTA_A")))
            break;

    assert(V = (Vertex *) calloc(1 + VSize, sizeof(Vertex)));
    for (i = 1; i <= VSize; i++) {
        v = &V[i];
        v->Id = i;
        v->InDegree = v->OutDegree = 0;
    }
    /* Create arcs */
    for (i = 0; i < RequiredEdges; i++) {
        int a, b, c, d = -1;
        fscanf(in, "(%d,%d)   coste  %d %d\n", &a, &b, &c, &d);
        v1 = &V[a];
        v2 = &V[b];
        if (c >= 0 && c != 9999 && c != 99999999) {
            v1->FirstArc = CreateArc(b, c, 1, v1->FirstArc);
            RequiredArcs++;
            v1->OutDegree++;
            v2->InDegree++;
        }
        if (d >= 0 && d != 9999 && d != 99999999) {
            v2->FirstArc = CreateArc(a, d, 1, v2->FirstArc);
            RequiredArcs++;
            v2->OutDegree++;
            v1->InDegree++;
        }
        if (c == d)
            Edges++;
    }
    fgets(Line, 256, in);
    for (i = 0; i < NotRequiredEdges; i++) {
        int a, b, c, d = -1;
        fscanf(in, "(%d,%d)   coste  %d %d\n", &a, &b, &c, &d);
        v1 = &V[a];
        v2 = &V[b];
        if (c >= 0 && c != 9999 && c != 99999999)
            v1->FirstArc = CreateArc(b, c, 0, v1->FirstArc);
        if (d >= 0 && d != 9999 && d != 99999999)
            v2->FirstArc = CreateArc(a, d, 0, v2->FirstArc);
    }
    fgets(Line, 256, in);
    for (i = 1; i <= VSize; i++) {
        v = &V[i];
        if (v->InDegree == 0 && v->OutDegree == 0) {
            v->Required = 1;
            RequiredVertices++;
        }
    }
    fclose(in);

    WSize = RequiredVertices + RequiredArcs;
    if (Class[0] == 'W')
        Clusters = WSize - RequiredEdges;
    else
        Clusters = WSize - Edges;

    printf("n = %d, ", WSize);
    printf("m = %d\n", Clusters);

    /* Open GTSP problem file */
    sprintf(Line, "GTSP/%s.gtsp", argv[2]);
    assert(out = fopen(Line, "w"));
    assert(W = (Wertex *) calloc(1 + WSize, sizeof(Wertex)));
    for (i = 1; i <= WSize; i++)
        W[i].Id = i;

    OldWSize = WSize;
    WSize = 0;
    for (i = 1; i <= VSize; i++) {
        v = &V[i];
        if (v->Required) {
            WSize++;
            W[WSize].From = W[WSize].To = v;
            W[WSize].Cost = 0;
        } else {
            for (a = v->FirstArc; a; a = a->Next) {
                if (a->Required) {
                    WSize++;
                    W[WSize].From = v;
                    W[WSize].To = &V[a->To];
                    W[WSize].Cost = a->Cost;
                }
            }
        }
    }
    if (WSize != OldWSize) {
        printf("ERROR: WSize: %d != OldWSize: %d\n", WSize, OldWSize);
        exit(1);
    }

    /* Calculate transformed costs */
    assert(C = (int **) malloc((WSize + 1) * sizeof(int *)));
    for (i = 1; i <= WSize; i++) {
        assert(C[i] = (int *) malloc((WSize + 1) * sizeof(int)));
        for (j = 1; j <= WSize; j++)
            C[i][j] = 999999;
    }
    MakeHeap(VSize);
    for (x = 1; x <= WSize; x++) {
        Dijkstra(W[x].To);
        for (y = 1; y <= WSize; y++)
            if (y != x)
                C[x][y] = W[y].From->Cost + W[y].Cost;
    }

    /* Create GTSP problem file */
    fprintf(out, "NAME : %s\n", argv[2]);
    fprintf(out, "TYPE : AGTSP\n");
    fprintf(out, "DIMENSION : %d\n", WSize);
    fprintf(out, "GTSP_SETS : %d\n", Clusters);
    fprintf(out, "EDGE_WEIGHT_TYPE : EXPLICIT\n");
    fprintf(out, "EDGE_WEIGHT_FORMAT : FULL_MATRIX\n");
    fprintf(out, "EDGE_WEIGHT_SECTION\n");
    for (i = 1; i <= WSize; i++) {
        for (j = 1; j <= WSize; j++)
            fprintf(out, "%d ", C[i][j]);
        fprintf(out, "\n");
    }
    fprintf(out, "GTSP_SET_SECTION\n");
    i = 0;
    for (x = 1; x <= WSize; x++) {
        w = &W[x];
        if (w->Id == 0)
            continue;
        i++;
        fprintf(out, "%d %d", i, x);
        for (y = x + 1; y <= WSize; y++) {
            if (w->From == W[y].To && w->To == W[y].From) {
                fprintf(out, " %d", y);
                w->Id = W[y].Id = 0;
                break;
            }
        }
        fprintf(out, " -1\n");
    }
    fprintf(out, "EOF\n");
    fclose(out);
    if (i != Clusters) {
        printf("ERROR: Clusters: %d != OldClusters: %d\n", i, Clusters);
        exit(1);
    }
    return 0;
}

static Arc *CreateArc(int To, int Cost, int Required, Arc * Next)
{
    Arc *a;
    assert(a = (Arc *) malloc(sizeof(Arc)));
    a->To = To;
    a->Cost = Cost;
    a->Required = Required;
    a->Next = Next;
    return a;
}

/* The Dijkstra function computes all shortest paths to a source vertex.
 * At return, Cost of each vertex holds the cost of the shortest path
 * from that vertex to the source vertex, and Dad references the dad for the
 * node in the shortest path tree.
 */

static void Dijkstra(Vertex * Source)
{
    Vertex *Blue = Source, *v;
    Arc *a;
    int i;

    Blue->Loc = 0;
    Blue->Cost = 0;
    Blue->Dad = 0;
    for (i = 1; i <= VSize; i++) {
        v = &V[i];
        if (v != Blue) {
            v->Dad = Blue;
            v->Cost = 999999;
            HeapLazyInsert(v);
        }
    }
    for (a = Blue->FirstArc; a; a = a->Next) {
        v = &V[a->To];
        v->Dad = Blue;
        v->Cost = a->Cost;
        HeapSiftUp(v);
    }
    while ((Blue = HeapDeleteMin())) {
        for (a = Blue->FirstArc; a; a = a->Next) {
            v = &V[a->To];
            if (!v->Loc || Blue->Cost + a->Cost >= v->Cost)
                continue;
            v->Dad = Blue;
            v->Cost = Blue->Cost + a->Cost;
            HeapSiftUp(v);
        }
    }
}
